/*	WebDownloader for X-Window
 *	Copyright (C) 1999 Koshelev Maxim
 *	This Program is free but not GPL!!! You can't modify it
 *	without agreement with autor. You can't distribute modified
 *	program but you can distribute unmodified program.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include "var.h"
#include "locstr.h"
#include "history.h"
#include "config.h"

void get_size_of_clist();

const char *CFG_FILE=".ntrc/config";
const char *CFG_DIR=".ntrc";

extern tHistory *UrlHistory;
extern tHistory *PathHistory;


void set_config(char *line) {
    char temp[MAX_LEN];
    int prom;
    if(*line!='\n' && *line!='#') {
        sscanf(line,"%s %i",temp,&prom);
        if (equal(temp,"max_threads")) {
            if (prom<=50 && prom>0) MAX_THREADS=prom;
            return;
        };
        if (equal(temp,"max_log")) {
            if (prom>0 && prom<500) MAX_LOG_LENGTH=prom;
            return;
        };
        if (equal(temp,"max_retries")) {
            if (prom>=0) MAX_RETRIES=prom;
            return;
        };
        if (equal(temp,"timeout")) {
            if (prom<500 && prom>0) TIME_OUT=prom;
            return;
        };
        if (equal(temp,"break_timeout")) {
            if (prom<=60 && prom>=0) RETRY_TIME_OUT=prom;
            return;
        };
        if (equal(temp,"optimize")) {
            if (prom==1 || prom==0) RECURSIVE_OPTIMIZE=prom;
            return;
        };
        if (equal(temp,"del_completed")) {
            if (prom==1 || prom==0) DELETE_COMPLETED=prom;
            return;
        };
        if (equal(temp,"del_fataled")) {
            if (prom==1 || prom==0) DELETE_FATAL=prom;
            return;
        };
        if (equal(temp,"nice_decs")) {
            NICE_DEC_DIGITALS=prom;
            return;
        };
        if (equal(temp,"time_format")) {
            TIME_FORMAT=prom;
            return;
        };
        if (equal(temp,"retry_if_noreget")) {
            if (prom==1 || prom==0) RETRY_IF_NOREGET=prom;
            return;
        };
        if (equal(temp,"sleeptime")) {
            if (prom<60 && prom>0) RETRY_TIME_OUT=prom;
            return;
        };
        if (equal(temp,"savepath")) {
            GLOBAL_SAVE_PATH=copy_string(line+strlen("savepath")+1);
            return;
        };
        if (equal(temp,"xposition")) {
            WINDOW_X_POSITION=prom;
            return;
        };
        if (equal(temp,"yposition")) {
            WINDOW_Y_POSITION=prom;
            return;
        };
        if (equal(temp,"windowwidth")) {
            WINDOW_WIDTH=prom;
            return;
        };
        if (equal(temp,"windowheight")) {
            WINDOW_HEIGHT=prom;
            return;
        };
        if (equal(temp,"clist_height")) {
            WINDOW_CLIST_HEIGHT=prom;
            return;
        };
    };
};

void read_config() {
    char *homedir=getenv("HOME");
    char *cfgpath=new char[strlen(homedir)+strlen(CFG_FILE)+2];
    cfgpath[0]=0;
    strcat(cfgpath,homedir);
    strcat(cfgpath,"/");
    strcat(cfgpath,CFG_FILE);
    int fd=open(cfgpath,O_RDONLY);
    if (fd>0) {
        char temp[MAX_LEN];
        char *cur=temp;
        while(read(fd,cur,1)) {
            while(*cur!='\n') {
                cur++;
                if (read(fd,cur,1)==0) break;
            };
            *cur=0;
            set_config(temp);
            cur=temp;
        };
        close(fd);
    } else {
        save_config();
        printf("Can't open cfg file at '%s'\n",cfgpath);
        printf("Use default cfg :))...\n");
    };
    /*
      printf("MAX_THREADS:%i\n",MAX_THREADS);
      printf("MAX_LOG_LENGTH:%i\n",MAX_LOG_LENGTH);
      printf("MAX_RETRIES:%i\n",MAX_RETRIES);
      printf("TIME_OUT:%i\n",TIME_OUT);
      printf("RETRY_TIME_OUT:%i\n",RETRY_TIME_OUT);
      printf("RECURSIVE_OPTIMIZE:%i\n",RECURSIVE_OPTIMIZE);
     */
    delete cfgpath;
    if (!UrlHistory)
        UrlHistory=new tHistory;
    if (!PathHistory)
        PathHistory=new tHistory;
    load_strlist(UrlHistory, ".ntrc/history1");
    load_strlist(PathHistory,".ntrc/history2");
};

void save_config() {
    char *homedir=getenv("HOME");
    char *cfgpath=new char[strlen(homedir)+strlen(CFG_FILE)+2];
    cfgpath[0]=0;
    strcat(cfgpath,homedir);
    strcat(cfgpath,"/");
    strcat(cfgpath,CFG_FILE);
    int fd=open(cfgpath,O_TRUNC | O_CREAT |O_RDWR,S_IRUSR | S_IWUSR);
    if (fd<0) {
        char *cfg_dir=new char[strlen(CFG_DIR)+strlen(homedir)+3];
        *cfg_dir=0;
        strcat(cfg_dir,homedir);
        strcat(cfg_dir,"/");
        strcat(cfg_dir,CFG_DIR);
        mkdir(cfg_dir,S_IRWXU| S_IXGRP|S_IRGRP);
        fd=open(cfgpath,O_TRUNC | O_CREAT |O_RDWR,S_IRUSR | S_IWUSR);
        delete cfg_dir;
    };
    if (fd>0) {
        char data[MAX_LEN];
        sprintf(data,"max_threads %i\n\n",MAX_THREADS);
        write(fd,data,strlen(data));
        sprintf(data,"max_log %i\n\n",MAX_LOG_LENGTH);
        write(fd,data,strlen(data));
        sprintf(data,"max_retries %i\n\n",MAX_RETRIES);
        write(fd,data,strlen(data));
        sprintf(data,"timeout %i\n\n",TIME_OUT);
        write(fd,data,strlen(data));
        sprintf(data,"break_timeout %i\n\n",RETRY_TIME_OUT);
        write(fd,data,strlen(data));
        sprintf(data,"sleeptime %i\n\n",RETRY_TIME_OUT);
        write(fd,data,strlen(data));
        sprintf(data,"optimize %i\n\n",RECURSIVE_OPTIMIZE);
        write(fd,data,strlen(data));
        sprintf(data,"retry_if_noreget %i\n\n",RETRY_IF_NOREGET);
        write(fd,data,strlen(data));
        sprintf(data,"del_completed %i\n\n",DELETE_COMPLETED);
        write(fd,data,strlen(data));
        sprintf(data,"del_fataled %i\n\n",DELETE_FATAL);
        write(fd,data,strlen(data));
        sprintf(data,"nice_decs %i\n\n",NICE_DEC_DIGITALS);
        write(fd,data,strlen(data));
        sprintf(data,"time_format %i\n\n",TIME_FORMAT);
        write(fd,data,strlen(data));
        sprintf(data,"savepath %s\n\n",GLOBAL_SAVE_PATH);
        write(fd,data,strlen(data));
        sprintf(data,"xposition %i\n\n",WINDOW_X_POSITION);
        write(fd,data,strlen(data));
        sprintf(data,"yposition %i\n\n",WINDOW_Y_POSITION);
        write(fd,data,strlen(data));
        sprintf(data,"windowwidth %i\n\n",WINDOW_WIDTH);
        write(fd,data,strlen(data));
        sprintf(data,"windowheight %i\n\n",WINDOW_HEIGHT);
        write(fd,data,strlen(data));
        get_size_of_clist();
        sprintf(data,"clist_height %i\n\n",WINDOW_CLIST_HEIGHT);
        write(fd,data,strlen(data));
        close(fd);
    } else {
        printf("Can't write cfgfile to:%s\n",cfgpath);
    };
    save_strlist(UrlHistory,".ntrc/history1");
    save_strlist(PathHistory,".ntrc/history2");
    delete cfgpath;
};

void save_strlist(tStringList *what,char *where) {
    if (!what || !where) return;
    char *homedir=getenv("HOME");
    char *path=sum_strings(homedir,"/",where);
    int fd=open(path,O_TRUNC | O_CREAT |O_RDWR,S_IRUSR | S_IWUSR);
    if (fd>0) {
        tString *tmp=what->first();
        while(tmp) {
            write(fd,tmp->body,strlen(tmp->body));
            write(fd,"\n",strlen("\n"));
            tmp=what->prev();
        };
        close(fd);
    };
    delete(path);
};

void load_strlist(tStringList *where,char *what) {
    if (!what || !where) return;
    char *homedir=getenv("HOME");
    char *path=sum_strings(homedir,"/",what);
    int fd=open(path,O_RDONLY);
    where->done();
    if (fd>0) {
        char temp[MAX_LEN];
        char *cur=temp;
        while(read(fd,cur,1)) {
            while(*cur!='\n') {
                cur++;
                if (read(fd,cur,1)==0) break;
            };
            *cur=0;
            if (strlen(temp))
                where->add(temp);
            cur=temp;
        };
        close(fd);
    };
    delete(path);
};
