/*	WebDownloader for X-Window
 *	Copyright (C) 1999 Koshelev Maxim
 *	This Program is free but not GPL!!! You can't modify it
 *	without agreement with autor. You can't distribute modified 
 *	program but you can distribute unmodified program.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

#include "ftpd.h"
#include "ftp.h"
#include "client.h"
#include "liststr.h"
#include "var.h"
#include "locstr.h"
#include "log.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>

//****************************************/
void extract_link(char *src,char *dst) {
    char *delim=" -> ";
    if (src) {
        char *tmp=strstr(src,delim);
        if (tmp) {
            strcpy(dst,tmp+strlen(delim));
            del_crlf(dst);
        } else dst[0]=0;
    };
};

int type_from_str(char *data) {
    if (index(data,'d'))
        return T_DIR;
    if (index(data,'l'))
        return T_LINK;
    // if 'c' or 'b' then T_DEVICE
    return T_FILE;
};

void cut_string_list(char *src,tFileInfo *dst,int flag) {
    char str1[MAX_LEN],name[MAX_LEN],attr[MAX_LEN];
    int par1;
    sscanf(src,"%s %u %s %s %u %s %u %s %s",
           attr,&par1,str1,str1,&dst->size,str1,&par1,str1,name);
    /* Next cycle extract name from string of
     * directory listing
     */
    dst->type=type_from_str(attr);
    char *tmp=src;
    for (int i=0;i<8;i++) {
        tmp=index(tmp,' ');
        if (tmp) while (*tmp==' ') tmp++;
        else break;
    };
    if (flag) {
        if (tmp) dst->name=copy_string(tmp);
        else dst->name=copy_string(name);
        del_crlf(dst->name);
    };
    if (dst->type==T_LINK) {
        extract_link(src,name);
        dst->body=copy_string(name);
        if (flag) {
            tmp=strstr(dst->name," -> ");
            if (tmp) *tmp=0;
        };
    } else dst->body=NULL;
};
//****************************************/

int tFtpDownload::change_dir() {
    if (CWDFlag) return 0;
    if (FTP->change_dir(D_PATH)) return -1;
    CWDFlag=1;
    return 0;
};

void tFtpDownload::print_reget(int offset) {
    if (!FTP->test_reget()) {
        LOG->add("Reget not supported!!!\n",LOG_WARNING);
    };
};

void tFtpDownload::check_for_repeated(tStringList *LIST) {
    if (!LIST) return;
    tString *temp,*point=NULL,*first;
    first=LIST->first();
    while((temp=LIST->prev())) {
        if (equal(temp->body,first->body)) {
            point=temp;
            break;
        };
    };
    if (point) {
        temp=LIST->first();
        while(temp && temp!=point) {
            LIST->del(temp);
            delete(temp);
            temp=LIST->first();
        };
    };
};
//****************************************/
tFtpDownload::tFtpDownload() {
    HOST=USER=PASS=D_PATH=NULL;
    D_FILE.name=D_FILE.body=NULL;
    D_FILE.size=D_FILE.type=D_FILE.fdesc=0;
    LOG=NULL;
    FTP=NULL;
    DIR=NULL;
    list=NULL;
    RetrNum=0;
    Status=D_NOTHING;
};

int tFtpDownload::reconnect() {
    int success=1;
    Status=D_QUERYING;
    while (success) {
        RetrNum++;
        char data[MAX_LEN];
        if (FTP->get_status()==STATUS_FATAL) return -1;
        if (FTP->get_status()!=STATUS_TIMEOUT && FTP->get_status()!=0) {
            LOG->add("Server refused login probably too many users of your class\n",LOG_WARNING);
        };
        if (config.number_of_attempts)
            sprintf(data,"Retrying %i of %i....\n",RetrNum,config.number_of_attempts);
        else
            sprintf(data,"Retrying %i ...\n",RetrNum);
        LOG->add(data,LOG_OK);
        if (RetrNum==config.number_of_attempts) {
            LOG->add("Max amount of retries was reached!\n",LOG_ERROR);
            return -1;
        };
        FTP->down();
        if (RetrNum>1) {
            if (FTP->test_reget() || RETRY_IF_NOREGET) {
                LOG->add("Sleeping\n",LOG_OK);
                sleep(config.time_for_sleep);
            } else return -1;
        };
        if (FTP->reinit()==0 && FTP->connect()==0) {
            success=0;
        };
    };
    CWDFlag=0;
    return 0;
};

int tFtpDownload::init(char *host,char *user,char *pass,int port,tLog *log,tCfg *cfg) {
    LOG=log;
    FTP=new tFtpClient;
    RetrNum=0;
    HOST=host;
    if (user)
        USER=user;
    else
        USER=DEFAULT_USER;
    if (pass)
        PASS=pass;
    else
        PASS=DEFAULT_PASS;
    DIR=NULL;
    list=NULL;
    D_FILE.fdesc=0;
    config.timeout=cfg->timeout;
    config.time_for_sleep=cfg->time_for_sleep;
    config.number_of_attempts=cfg->number_of_attempts;
    FTP->init(host,LOG,port,config.timeout);
    FTP->registr(USER,PASS);
    return reconnect();
};

tStringList *tFtpDownload::dir() {
    return DIR;
};

int tFtpDownload::get_size() {
    if (!list) {
        list=new tStringList;
        list->init(0);
    } else list->done();
    int needcheck=0;
    while (1) {
        if (!change_dir()) {
            if (!FTP->stand_data_connection()) {
                int a=FTP->get_size(D_FILE.name,list);
                if (a==0 && list->count()<=2) {
                    tString *last=list->last();
                    if (!last) {
                        D_FILE.type=T_FILE;
                        D_FILE.size=0;
                        D_FILE.body=NULL;
                        return 0;
                    };
                    if (equal_first(last->body,"total")) {
                        D_FILE.type=T_DIR;
                        D_FILE.size=1;
                        DIR=list;
                        list=NULL;
                        D_FILE.body=NULL;
                        return 0;
                    };
                    char str1[MAX_LEN];
                    tFileInfo temp;
                    cut_string_list(last->body,&temp,0);
                    D_FILE.type=temp.type;
                    D_FILE.body=temp.body;
                    D_FILE.size=temp.size;
                    sprintf(str1,"Length is %i\n",D_FILE.size);
                    LOG->add(str1,LOG_OK);
                    return D_FILE.size;
                };
                if (a==0 && list->count()>2) {
                    LOG->add("This is a directory!\n",LOG_WARNING);
                    D_FILE.size=1;
                    if (DIR) delete DIR;
                    DIR=list;
                    list=NULL;
                    D_FILE.type=T_DIR;
                    if (needcheck) check_for_repeated(DIR);
                    return 1;
                };
                needcheck=1;
                LOG->add("Couldn't get size :((\n",LOG_WARNING);
            } else {
                LOG->add("Can't establish data connection\n",LOG_ERROR);
            };
        } else {
            int s=FTP->get_status();
            if (s!=STATUS_TIMEOUT && s!=STATUS_CMD_ERR)
                break;
            LOG->add("Can't change dir!\n",LOG_ERROR);
        };
        if (reconnect())
            break;
    };
    return -1;
};

int tFtpDownload::download_dir() {
    unsigned int offset=0;
    int ind=0,needcheck=0;
    if (!DIR) {
        DIR=new tStringList;
        DIR->init(0);

        while(1) {
            if (!change_dir()) {
                if (!FTP->stand_data_connection()) {
                    print_reget(offset);
                    if (!FTP->test_reget()) {
                        offset=0;
                        DIR->done();
                    };
                    Status=D_DOWNLOAD;
                    ind=FTP->get_size(D_FILE.name,DIR);
                    if (ind==0) {
                        LOG->add("Listing was loaded\n",LOG_OK);
                        if (needcheck) check_for_repeated(DIR);
                        return 0;
                    };
                    LOG->add("Listing not loaded completelly\n",LOG_WARNING);
                    needcheck=1;
                } else {
                    LOG->add("Can't establish data connection\n",LOG_ERROR);
                };
            } else {
                LOG->add("Can't change directory\n",LOG_ERROR);
                if (FTP->get_status()==STATUS_FATAL)
                    return -1;
            };
            if (reconnect()) return -1;
        };
    };
    return 0;
};




int tFtpDownload::download(unsigned int from,unsigned int len) {
    unsigned int length=0,offset=from;
    int ind=0;
    if (D_FILE.type==T_DIR) {
        LOG->add("Loading directory...\n",LOG_OK);
        return(download_dir());
    };
    if (D_FILE.type==T_LINK) {
        LOG->add("Link was loading :))\n",LOG_OK);
        return 0;
    };
    while(1) {
        if (!change_dir()) {
            if (!FTP->stand_data_connection()) {
                print_reget(from);
                if (!FTP->test_reget())
                    StartSize=offset=0;
                if (lseek(D_FILE.fdesc,offset,SEEK_SET)<0) {
                    LOG->add("Problems with lseek()\n",LOG_WARNING);
                };
                Status=D_DOWNLOAD;
                ind=FTP->get_file_from(D_FILE.name,offset,D_FILE.fdesc);
                if (ind>0) {
                    char data[MAX_LEN];
                    length+=ind;
                    offset+=ind;
                    sprintf(data,"%i bytes loaded.\n",length);
                    LOG->add(data,LOG_OK);
                };
                if (!FTP->get_status()) return 0;
            } else {
                LOG->add("Can't establish data connection\n",LOG_ERROR);
            };
        } else {
            LOG->add("Can't change directory\n",LOG_ERROR);
            if (FTP->get_status()!=STATUS_TIMEOUT)
                return -1;
        };
        if (reconnect()) return -1;
    };
    return 0;
};

int tFtpDownload::get_readed() {
    return (FTP->get_readed());
};

int tFtpDownload::get_child_status() {
    return(FTP->get_status());
};

void tFtpDownload::done() {
    if (FTP) FTP->done();
    if (D_FILE.fdesc) {
        close(D_FILE.fdesc);
        D_FILE.fdesc=0;
    };
};

tFtpDownload::~tFtpDownload() {
    if (FTP) delete(FTP);
    if (D_PATH) delete(D_PATH);
    if (D_FILE.name) delete(D_FILE.name);
    if (D_FILE.body) delete D_FILE.body;
    if (D_FILE.fdesc) close(D_FILE.fdesc);
    if (DIR) delete(DIR);
    if (list) delete(list);
};
