/*	WebDownloader for X-Window
 *	Copyright (C) 1999 Koshelev Maxim
 *	This Program is free but not GPL!!! You can't modify it
 *	without agreement with autor. You can't distribute modified
 *	program but you can distribute unmodified program.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <pthread.h>
#include <time.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <errno.h>

#include "var.h"
#include "initvar.cc"
#include "ftpd.h"
#include "httpd.h"
#include "locstr.h"
#include "main.h"
#include "dlist.h"
#include "mdlist.h"
#include "meter.h"
#include "log.h"
#include "mainlog.h"
#include "signal.h"
#include "savelog.h"
#include "face/list.h"
#include "face/log.h"

tTwoStrings::tTwoStrings() {
    one=two=NULL;
};

void tTwoStrings::zero() {
    one=two=NULL;
};

tTwoStrings::~tTwoStrings() {};

int get_port_by_proto(char *proto) {
    if (proto) {
        if (equal(proto,"ftp")) return 21;
        if (equal(proto,"http")) return 80;
    };
    return 21;
};

//**********************************************/


tDList *RunList,*StopList,*CompleteList,*PausedList;
tDList *WaitList;
tMeter *GlobalMeter=NULL;
key_t LogsMsgQueue;

void tMain::init() {
    WaitList=new tDList(DL_WAIT);
    RunList=new tDList(DL_RUN);
    StopList=new tDList(DL_STOP);
    PausedList=new tDList(DL_PAUSE);
    CompleteList=new tDList(DL_COMPLETE);
    WaitList->init(0);
    RunList->init(0);
    StopList->init(0);
    PausedList->init(0);
    CompleteList->init(0);
    GlobalMeter=new tMeter;
    GlobalMeter->init(METER_LENGTH);
    LastReadedBytes=0;
    /* Create msgqueue for logs update */
    LogsMsgQueue=100;
    while ((MsgQueue=msgget(LogsMsgQueue,IPC_CREAT|0600))<0) {
        LogsMsgQueue+=1;
    };
    /* Clearing msgqueue */
    int complete=0;
    while (!complete) {
        mbuf Msg;
        complete=1;
        if (msgrcv(MsgQueue,(msgbuf *)&Msg,sizeof(Msg)-sizeof(long),1,IPC_NOWAIT)>0 ||
                    msgrcv(MsgQueue,(msgbuf *)&Msg,sizeof(Msg)-sizeof(long),2,IPC_NOWAIT)>0)
            complete=0;
    };
};

void tMain::load_defaults() {
    tStringList *list=new tStringList;
    list->init(0);
    read_list(list);
    tString *temp=list->last();
    while(temp) {
        if (strlen(temp->body)) {
            tString *temp2=list->next();
            add_downloading(temp2->body,temp->body);
        };
        temp=list->next();
    };
    delete(list);
};

void tMain::init_main_log(){
	MainLog=new tMLog;
    MainLog->init_list(GTK_CLIST(MainLogList));
    MainLog->init(MAX_LOG_LENGTH);
    MainLog->add(VERSION_NAME,LOG_WARNING);
    MainLog->add("Normaly started",LOG_WARNING);
};

void tMain::redraw_logs() {
    mbuf Msg;
    int complete=0;
    int limit = 0;
    while (!complete && limit<100) {
        complete=1;
        limit++;
        if (msgrcv(MsgQueue,(msgbuf *)&Msg,sizeof(Msg)-sizeof(long),1,IPC_NOWAIT)>0) {
            complete=0;
            log_window_add_string(Msg.which,Msg.what);
        };
        if (msgrcv(MsgQueue,(msgbuf *)&Msg,sizeof(Msg)-sizeof(long),2,IPC_NOWAIT)>0) {
            complete=0;
            del_first_from_log(Msg.which);
        };
    };
};

void tMain::split_string(char *what,char *delim,tTwoStrings *out) {
    char * where=strstr(what,delim);
    if (where) {
        int len=strlen(where),len1=strlen(delim);
        out->two=copy_string(where+len1);
        len1=strlen(what)-len;
        out->one=copy_string(what,len1);
    } else {
        out->two=copy_string(what);
        out->one=NULL;
    };
    delete(what);
};

void tMain::del_completed() {
    tDownload *temp=CompleteList->last();
    while (temp) {
        del_download_from_clist(temp);
        CompleteList->del(temp);
        delete temp;
        temp=CompleteList->last();
    };
    DEL_COMPLETED=0;
};

void tMain::del_fataled() {
    tDownload *temp=StopList->last();
    while (temp) {
        del_download_from_clist(temp);
        StopList->del(temp);
        delete temp;
        temp=StopList->last();
    };
    DEL_FATALED=0;
};

void tMain::del_all() {
	MainLog->add("Clear queue of downloads",LOG_ERROR);
    del_completed();
    del_fataled();
    tDownload *temp=PausedList->last();
    while (temp) {
        del_download_from_clist(temp);
        PausedList->del(temp);
        delete temp;
        temp=PausedList->last();
    };
    temp=WaitList->last();
    while (temp) {
        del_download_from_clist(temp);
        WaitList->del(temp);
        delete temp;
        temp=WaitList->last();
    };
    temp=RunList->last();
    while (temp) {
        stop_thread(temp);
        del_download_from_clist(temp);
        RunList->del(temp);
        delete temp;
        temp=RunList->last();
    };
};

tAddr *tMain::analize(char *what) {
    tTwoStrings pair;
    split_string(what,"://",&pair);
    tAddr *out=new tAddr;
    if (pair.one) {
        out->protocol=pair.one;
    } else {
        out->protocol=copy_string(DEFAULT_PROTO);
    };
    out->host=pair.two;
    if (!out->host) {
        delete(out);
        return NULL;
    };
    split_string(out->host,"/",&pair);
    if (pair.one) {
        out->host=pair.one;
        out->file=pair.two;
    } else {
        out->host=pair.two;
        out->file=pair.one;
    };
    split_string(out->host,"@",&pair);
    out->host=pair.two;
    out->username=pair.one;
    if (out->username) {
        split_string(out->username,":",&pair);
        out->username=pair.one;
        out->pass=pair.two;
    } else {
        out->username=NULL;
        out->pass=NULL;
    };
    if (out->file) {
        char *tmp=parse_percents(out->file);
        if (tmp && equal(out->protocol,"ftp")) {
            delete out->file;
            out->file=tmp;
        };
        if (out->file[strlen(out->file)-1]=='/')
            out->file[strlen(out->file)-1]=0;
        char *prom=rindex(out->file,'/');
        if (prom) {
            out->path=copy_string(prom+1);
            *prom=0;
            prom=out->path;
            out->path=sum_strings("/",out->file);
            delete out->file;
            out->file=prom;
        };
    } else {
        out->file=copy_string("");
    };
    if (!out->path) out->path=copy_string("/");
    split_string(out->host,":",&pair);
    if (pair.one) {
        sscanf(pair.two,"%i",&out->port);
        delete pair.two;
        out->host=pair.one;
    } else {
        out->port=0;
        out->host=pair.two;
    };
    /* Parse # in http urls
     */
    if (equal(out->protocol,"http") && out->file!=NULL) {
        char *tmp=index(out->file,'#');
        if (tmp) {
            *tmp=0;
            tmp=out->file;
            out->file=copy_string(tmp);
            delete(tmp);
        };
    };
    return out;
};

void tMain::run_new_thread(tDownload *what) {
    pthread_attr_t attr_p;
    what->status=READY_TO_RUN;
    if (!what->LOG) {
        what->LOG=new tLog;
        what->LOG->init(MAX_LOG_LENGTH);
    };
    what->delete_editor();
    pthread_attr_init(&attr_p);
    pthread_attr_setdetachstate(&attr_p,PTHREAD_CREATE_JOINABLE);
    pthread_attr_setscope(&attr_p,PTHREAD_SCOPE_SYSTEM);
    pthread_create(&what->thread_id,&attr_p,download_last,(void *)what);
};

void tMain::stop_download(tDownload *what) {
    if (!RunList->owner(what)) return;
    stop_thread(what);
    RunList->del(what);
    what->LOG->add("Download  was stopped by user",LOG_WARNING);
    char data[MAX_LEN];
    sprintf(data,"Downloaing of file %s from %s was stopped by user",what->info->file,what->info->host);
    MainLog->add(data,LOG_WARNING);
    PausedList->insert(what);
    set_pixmap_pause(what->GTKCListRow);
    what->Status.clear();
};

void tMain::delete_download(tDownload *what) {
	char data[MAX_LEN];
	sprintf(data,"Delete file %s from queue of downloads",what->info->file);
	MainLog->add(data,LOG_WARNING);
    if (!what) return;
    switch (what->owner) {
        case DL_STOP: {
                StopList->del(what);
                break;
            };
        case DL_COMPLETE: {
                CompleteList->del(what);
                break;
            };
        case DL_RUN: {
                stop_thread(what);
                RunList->del(what);
                break;
            };
        case DL_PAUSE: {
                PausedList->del(what);
                break;
            };
        case DL_WAIT: WaitList->del(what);
    };
    del_download_from_clist(what);
    delete (what);
};

void tMain::continue_download(tDownload *what) {
    if (!PausedList->owner(what) && !StopList->owner(what)) return;
    if (PausedList->owner(what))
        PausedList->del(what);
    if (StopList->owner(what))
        StopList->del(what);
    if (RunList->count()<MAX_THREADS) {
        RunList->insert(what);
        set_pixmap_part_run(what->GTKCListRow);
        run_new_thread(what);
    } else {
        tDownload *temp=WaitList->last();
        if (!temp || temp->GTKCListRow < what->GTKCListRow)
            WaitList->insert(what);
        else {
        	temp=WaitList->first();
        	while (temp && temp->GTKCListRow < what->GTKCListRow)
        		temp=WaitList->prev();
        	if (temp) WaitList->insert_before(what,temp);
        	else{
        		temp=WaitList->first();
        		if (temp->GTKCListRow < what->GTKCListRow)
        			WaitList->insert(what);
        		else
        			WaitList->insert_before(what,temp);
        	};
        };
        set_pixmap_wait(what->GTKCListRow);
    };
    char data[MAX_LEN];
    sprintf(data,"Continue downloading of file %s from %s...",what->info->file,what->info->host);
    MainLog->add(data,LOG_OK);
    what->Attempt.clear();
};

int tMain::complete() {
    return (WaitList->count()+RunList->count())>0?0:1;
};

void tMain::add_dir(tStringList *dir,tDownload *parent) {
    if (!dir) return;
    tAddr *addr=parent->info;
    tString *temp=dir->first();
    if (!temp) return;
    if (equal_first(temp->body,"total")) {
        dir->del(temp);
        delete temp;
    };
    char *path,*SavePath;
    int len=strlen(addr->path);
    /* There are we make new file's path
     */
    if (len && addr->path[len-1]!='/' && addr->file[0]!='/')
        path=sum_strings(addr->path,"/",addr->file);
    else
        path=sum_strings(addr->path,addr->file);
    /* There are we make new path where files will be saved
     */
    if (parent->SavePath) {
        len=strlen(parent->SavePath);
        if (len && parent->SavePath[len-1]!='/' && addr->file[0]!='/')
            SavePath=sum_strings(parent->SavePath,"/",addr->file);
        else
            SavePath=sum_strings(parent->SavePath,addr->file);
    } else {
        SavePath=copy_string(addr->file);
    };
    temp=dir->last();
    while (temp) {
        tFileInfo prom;
        prom.name=NULL;
        prom.body=NULL;
        cut_string_list(temp->body,&prom,1);
        if (!equal(prom.name,".") && !equal(prom.name,"..")) {
            tAddr *addrnew=new tAddr;
            addrnew->protocol=copy_string(addr->protocol);
            addrnew->host=copy_string(addr->host);
            addrnew->path=copy_string(path);
            addrnew->port=addr->port;
            addrnew->file=prom.name;

            addrnew->pass=copy_string(addr->pass);
            addrnew->username=copy_string(addr->username);

            tDownload *onenew=new tDownload;

            onenew->info=addrnew;
            onenew->SavePath=copy_string(SavePath);

            onenew->config.timeout=parent->config.timeout;
            onenew->config.time_for_sleep=parent->config.time_for_sleep;
            onenew->config.number_of_attempts=parent->config.number_of_attempts;

            if (RECURSIVE_OPTIMIZE) {
                onenew->FileType=prom.type;
                onenew->FullSize=prom.size;
                if (onenew->FileType==T_LINK) {
                    onenew->FileBody=prom.body;
                };
            } else {
                if (prom.body) delete prom.body;
            };
            WaitList->insert(onenew);
            add_download_to_clist(onenew);
        } else {
            if (prom.name) delete(prom.name);
            if (prom.body) delete(prom.body);
        };
        temp=dir->next();
    };
    delete (path);
    delete (SavePath);
};

void tMain::print_info(tDownload *what) {
    char data[MAX_LEN];
    switch(what->FileType) {
        case T_FILE: {
                change_clist_data(what->GTKCListRow,FILE_TYPE_COL,"file");
                make_number_nice(data,what->FullSize);
                change_clist_data(what->GTKCListRow,FULL_SIZE_COL,data);
                if (what->who) what->Size.set(what->who->get_readed());
                if (what->Size.change()){
                	make_number_nice(data,what->Size.curent);
                	what->Size.reset();
                	change_clist_data(what->GTKCListRow,DOWNLOADED_SIZE_COL,data);
                	what->Pause=time(NULL);
                   	change_clist_data(what->GTKCListRow,PAUSE_COL,"");
                }else{
               		if (what->status==DOWNLOAD_GO) {
                   		int Pause=time(NULL)-what->Pause;
                   		if (Pause>=30) {
                       		convert_time(Pause,data);
                       		change_clist_data(what->GTKCListRow,PAUSE_COL,data);
                    	};
                	};
                };
                if (what->Start>0) {
                    convert_time(time(NULL)-what->Start,data);
                    change_clist_data(what->GTKCListRow,TIME_COL,data);
                };
                int temp;
                if (what->FullSize)
                    temp=int((float(what->Size.curent)/float(what->FullSize))*100);
                else
                    temp=100;
                what->Percent.set(temp);
                sprintf(data,"%i",temp);
                if (what->Percent.change()){
                	change_clist_data(what->GTKCListRow,PERCENT_COL,data);
                	what->Percent.reset();
                };
                /*	Speed calculation
                 */
                int period=int(time(NULL)-what->Start);
                if (period && what->who) {
                    float tmp=float(what->Size.curent-what->who->get_start_size());
                    int speed=0;
                    if (tmp>0) {
                        speed=int(tmp/period);
                        sprintf(data,"%i",speed);
                        change_clist_data(what->GTKCListRow,SPEED_COL,data);
                    };
                    /*	Remaining time calculation
                     */
                    if (tmp>0) {
                        tmp=(float(what->FullSize-what->Size.curent)*float(period))/tmp;
                        if (tmp>=0 && tmp<24*3660) {
                            convert_time(int(tmp),data);
                            change_clist_data(what->GTKCListRow,ELAPSED_TIME_COL,data);
                        } else
                            change_clist_data(what->GTKCListRow,ELAPSED_TIME_COL,"...");
                    } else
                        change_clist_data(what->GTKCListRow,ELAPSED_TIME_COL,"...");
                };
                break;
            };
        case T_DIR: {
                change_clist_data(what->GTKCListRow,FILE_TYPE_COL,"dir");
                break;
            };
        case T_LINK: {
                change_clist_data(what->GTKCListRow,FILE_TYPE_COL,"link");
                break;
            };
        default:
            change_clist_data(what->GTKCListRow,FILE_TYPE_COL,"???");
    };
    if (what->who) {
    	what->Status.set(what->who->get_status());
    	if (what->Status.change()){
    		what->Status.reset();
    		switch (what->Status.curent){
    			case D_QUERYING:{
    				set_pixmap_part_run(what->GTKCListRow);
    				break;
    			};
    			case D_DOWNLOAD:{
    				set_pixmap_run(what->GTKCListRow);
    				break;
    			};
    		};
    	};
    	what->Attempt.set(what->who->treat());
    	if (what->Attempt.change()){
    		what->Attempt.reset();
        	if (what->config.number_of_attempts > 0)
            	sprintf(data,"%i/%i",what->Attempt.curent, what->config.number_of_attempts);
        	else
            	sprintf(data,"%i",what->Attempt.curent);
	        change_clist_data(what->GTKCListRow,TREAT_COL,data);
        };
    };
};


void tMain::main_circle() {
    if (DEL_COMPLETED) del_completed();
    if (DEL_FATALED) del_fataled();
    tDownload *temp=RunList->last();
    freeze_clist();
    while(temp) {
        tDownload *temp1=RunList->next();
        print_info(temp);
       	char data[MAX_LEN];
        switch(temp->status) {
            case DOWNLOAD_COMPLETE: {
                    RunList->del(temp);
                    if (temp->who->file_type()==T_DIR) {
                    	sprintf(data,"Downloading of directory %s from %s was completed",temp->info->file,temp->info->host);
                        tFtpDownload *prom=(tFtpDownload *)(temp->who);
                        add_dir(prom->dir(),temp);
                    } else{
                    	sprintf(data,"Downloading of file %s from %s was completed",temp->info->file,temp->info->host);
                    	make_file_visible(temp);
                    };
                   	MainLog->add(data,LOG_OK);
                    if (temp->FileType==T_REDIRECT) {
                        char newurl[MAX_LEN];
                        if (temp->who->get_new_url(newurl)==0) {
                            add_downloading(newurl,temp->SavePath);
                        };
                    };
//                    temp->who->done();
                    if (DELETE_COMPLETED) {
                        del_download_from_clist(temp);
                        delete (temp);
                    } else {
                        delete(temp->who);
                        temp->Percent.clear();
                        temp->who=NULL;
                        CompleteList->insert(temp);
                        set_pixmap_complete(temp->GTKCListRow);
                    };
                    break;
                };
            case DOWNLOAD_FATAL: {
                   	sprintf(data,"Downloading of file %s from %s was breaked by fatal error",temp->info->file,temp->info->host);
                   	MainLog->add(data,LOG_ERROR);
                    RunList->del(temp);
//                    temp->who->done();
                    if (DELETE_FATAL) {
                        del_download_from_clist(temp);
                        delete (temp);
                    } else {
                        delete(temp->who);
                        temp->Percent.clear();
                        temp->who=NULL;
                        StopList->insert(temp);
                        set_pixmap_stop(temp->GTKCListRow);
                    };
                };
        };
        temp=temp1;
    };
    temp=WaitList->first();
    while(temp && RunList->count()<MAX_THREADS) {
        WaitList->del(temp);
        RunList->insert(temp);
        set_pixmap_part_run(temp->GTKCListRow);
        run_new_thread(temp);
        temp=WaitList->first();
    };
    unfreeze_clist();
    speed();
};

void tMain::make_file_visible(tDownload *what) {
    if (what->FileType==T_FILE) {
        char *savepath=what->SavePath;
        char *oldname=sum_strings(savepath,"/.",what->who->get_real_name());
        char *newname=sum_strings(savepath,"/",what->who->get_real_name());
        rename(oldname,newname);
        delete oldname;
        delete newname;
    };
};

//**********************************************/

int tMain::add_downloading(char *adr,char *where) {
    char *temp=copy_string(adr);
    tAddr *addr=analize(temp);
    if (!addr) return -1;
    if (!addr->port)
        addr->port=get_port_by_proto(addr->protocol);
    tDownload *whatadd=new tDownload;
    whatadd->info=addr;
    if (where!=NULL && strlen(where)>0)
        whatadd->SavePath=copy_string(where);
    if (strlen(addr->file)==0) {
        whatadd->FileType=T_DIR;
        whatadd->FullSize=0;
        char *NewPath=sum_strings(whatadd->SavePath,"/",addr->host);
        if (mkdir(NewPath,S_IRWXU)==0 || errno==EEXIST) {
            delete (whatadd->SavePath);
            whatadd->SavePath=NewPath;
        } else
            delete(NewPath);
    };
    whatadd->config.timeout=TIME_OUT;
    whatadd->config.time_for_sleep=RETRY_TIME_OUT;
    whatadd->config.number_of_attempts=MAX_RETRIES;

    WaitList->insert(whatadd);
    add_download_to_clist(whatadd);
    addr=NULL;
    return 0;
};


void tMain::speed() {
    time_t TimeLeft=time(NULL)-LastTime;
    if (TimeLeft>0) {
        int Speed=(READED_BYTES-LastReadedBytes)/TimeLeft;
        LastReadedBytes=READED_BYTES;
        GlobalMeter->add(Speed);
        LastTime=time(NULL);
    };
};

void tMain::run() {
    LastTime=time(NULL);
    gtk_main();
};

void tMain::add_download_message(tDownload *what){
	if (!what) return;
	char data[MAX_LEN];
	sprintf(data,"Added downloading of file %s from %s [by user]",what->info->file,what->info->host);
	MainLog->add(data,LOG_OK);
};

void tMain::done() {
    /* There are  we MUST stop all threads!!!
     */
    tDownload *tmp=RunList->last();
    while (tmp) {
        stop_download(tmp);
        tmp=RunList->next();
    };
    delete(WaitList);
    delete(RunList);
    delete(StopList);
    delete(CompleteList);
    delete(GlobalMeter);
    msgctl(MsgQueue,IPC_RMID,NULL);
};
//*****************************************************************/
void download_http(tDownload *what) {
    if (!what->who) what->who=new tHttpDownload;
    tAddr *addr=what->info;
    if (what->who->init(addr->host,addr->username,addr->pass,addr->port,what->LOG,&(what->config))) {
        what->status=DOWNLOAD_FATAL;
        what->who->done();
        pthread_exit(NULL);
        return;
    };
    what->who->init_download(addr->path,addr->file);
    /* We need to know size of already loaded file
     * but I think if file not found we need to delete it
     * because in http name of file may be specify 
     * in http answer
     */
    int CurentSize=what->who->create_file(what->SavePath);
    if (CurentSize<0) {
        what->status=DOWNLOAD_FATAL;
        what->who->done();
        pthread_exit(NULL);
        return;
    };
    what->who->set_data(CurentSize);
    int size=what->who->get_size();
    /* In the case if file already loaded
     */
    if (size==CurentSize && size!=0) {
        what->FullSize=size;
        what->FileType=T_FILE;
        what->who->done();
        what->status=DOWNLOAD_COMPLETE;
        pthread_exit(NULL);
        return;
    };
    /* There are must be procedure of removing file
     * wich execute if CurentSize==0
     */
    if (CurentSize==0) {
        if (what->who->delete_file(what->SavePath))
            what->LOG->add("It is strange that we can't delete file which just created...",LOG_WARNING);
    };

    if (size<-1) {
        what->LOG->add("File not found\n",LOG_WARNING);
        what->who->done();
        what->status=DOWNLOAD_FATAL;
        pthread_exit(NULL);
        return;
    };
    if (size==-1) {
        what->FileType=T_REDIRECT;
        what->who->done();
        what->status=DOWNLOAD_COMPLETE;
        pthread_exit(NULL);
        return;
    };
    what->FullSize=size;
    what->FileType=T_FILE;
    what->Start=what->Pause=time(NULL);
    /* there we need to create file again
     * if CurentSize==0
     */
    if (CurentSize==0) CurentSize=what->who->create_file(what->SavePath);
    what->status=DOWNLOAD_GO;
    if (what->who->download(CurentSize,what->FullSize-CurentSize)) {
        what->who->done();
        what->status=DOWNLOAD_FATAL;
        pthread_exit(NULL);
        return;
    };
    what->who->done();
    what->status=DOWNLOAD_COMPLETE;
};

//-----------------------------------------------------------------

void *download_last(void *nothing) {
    init_signal_handler();
    tDownload *what=(tDownload *)nothing;
    if (what) {
        tAddr *addr=what->info;
        what->LOG->MsgQueue=msgget(LogsMsgQueue,0);
        if (what->LOG->MsgQueue<0) {
            what->LOG->add("Can't open messages queue!\n",LOG_ERROR);
            what->status=DOWNLOAD_FATAL;
            pthread_exit(NULL);
            return NULL;
        };
        if (addr==NULL || equal_first(addr->protocol,"ftp")) {
            if (!what->who) what->who=new tFtpDownload;
        } else {
            download_http(what);
            pthread_exit(NULL);
            return NULL;
        };
        what->LOG->add("Was Started!\n",LOG_WARNING);

        if (what->FileType==T_LINK) {
            what->LOG->add("It is a link and we already load it\n",LOG_WARNING);
            what->who->short_init(what->LOG);
            what->who->init_download(addr->path,addr->file);
            what->who->set_file_type(what->FileType);
            what->who->set_file_body(what->FileBody);
            what->who->create_file(what->SavePath);
   	        what->who->done();
	        what->status=DOWNLOAD_COMPLETE;
            pthread_exit(NULL);
            return NULL;
        };

        if (what->who->init(addr->host,addr->username,addr->pass,addr->port,what->LOG,&(what->config))) {
	        what->who->done();
            what->status=DOWNLOAD_FATAL;
            pthread_exit(NULL);
            return NULL;
        };

        what->who->init_download(addr->path,addr->file);
        if (what->FullSize<0) {
            what->status=DOWNLOAD_SIZE_WAIT;
            int size=what->who->get_size();
            if (size<0) {
                what->LOG->add("File not found\n",LOG_ERROR);
		        what->who->done();
                what->status=DOWNLOAD_FATAL;
                pthread_exit(NULL);
                return NULL;
            };
            what->FullSize=size;
            what->FileType=what->who->file_type();
        } else {
            what->who->set_file_type(what->FileType);
            if (what->FileType==T_LINK)
                what->who->set_file_body(what->FileBody);
        };

        int CurentSize=what->who->create_file(what->SavePath);
        if (CurentSize<0) {
	        what->who->done();
            what->status=DOWNLOAD_FATAL;
            pthread_exit(NULL);
            return NULL;
        };

        what->status=DOWNLOAD_GO;
        what->Start=what->Pause=time(NULL);
        if (what->who->download(CurentSize,what->FullSize-CurentSize)) {
	        what->who->done();
            what->status=DOWNLOAD_FATAL;
            pthread_exit(NULL);
            return NULL;
        };
        what->who->done();
        what->status=DOWNLOAD_COMPLETE;
    };
    pthread_exit(NULL);
    return NULL;
};

